@@ -25,45 +25,64 @@ import inspect
 import keyword
 import pydoc
 import re
-from collections import namedtuple
-from typing import Any, Optional, Type
+from dataclasses import dataclass
+from typing import Any, Callable, Optional, Type, Dict, List, ContextManager
 from types import MemberDescriptorType, TracebackType
-from typing_extensions import Literal
+from ._typing_compat import Literal
 
 from pygments.token import Token
 from pygments.lexers import Python3Lexer
 
 from .lazyre import LazyReCompile
 
-ArgSpec = namedtuple(
-    "ArgSpec",
-    [
-        "args",
-        "varargs",
-        "varkwargs",
-        "defaults",
-        "kwonly",
-        "kwonly_defaults",
-        "annotations",
-    ],
-)
 
-FuncProps = namedtuple("FuncProps", ["func", "argspec", "is_bound_method"])
+class _Repr:
+    """
+    Helper for `ArgSpec`: Returns the given value in `__repr__()`.
+    """
+
+    __slots__ = ("value",)
+
+    def __init__(self, value: str) -> None:
+        self.value = value
+
+    def __repr__(self) -> str:
+        return self.value
+
+    __str__ = __repr__
 
 
-class AttrCleaner:
+@dataclass
+class ArgSpec:
+    args: List[str]
+    varargs: Optional[str]
+    varkwargs: Optional[str]
+    defaults: Optional[List[_Repr]]
+    kwonly: List[str]
+    kwonly_defaults: Optional[Dict[str, _Repr]]
+    annotations: Optional[Dict[str, Any]]
+
+
+@dataclass
+class FuncProps:
+    func: str
+    argspec: ArgSpec
+    is_bound_method: bool
+
+
+class AttrCleaner(ContextManager[None]):
     """A context manager that tries to make an object not exhibit side-effects
-    on attribute lookup."""
+    on attribute lookup.
+
+    Unless explicitly required, prefer `getattr_safe`."""
 
     def __init__(self, obj: Any) -> None:
-        self.obj = obj
+        self._obj = obj
 
-    def __enter__(self):
+    def __enter__(self) -> None:
         """Try to make an object not exhibit side-effects on attribute
         lookup."""
-        type_ = type(self.obj)
-        __getattribute__ = None
-        __getattr__ = None
+        type_ = type(self._obj)
         # Dark magic:
         # If __getattribute__ doesn't exist on the class and __getattr__ does
         # then __getattr__ will be called when doing
@@ -86,7 +105,7 @@ class AttrCleaner:
             except TypeError:
                 # XXX: This happens for e.g. built-in types
                 __getattribute__ = None
-        self.attribs = (__getattribute__, __getattr__)
+        self._attribs = (__getattribute__, __getattr__)
         # /Dark magic
 
     def __exit__(
@@ -96,8 +115,8 @@ class AttrCleaner:
         exc_tb: Optional[TracebackType],
     ) -> Literal[False]:
         """Restore an object's magic methods."""
-        type_ = type(self.obj)
-        __getattribute__, __getattr__ = self.attribs
+        type_ = type(self._obj)
+        __getattribute__, __getattr__ = self._attribs
         # Dark magic:
         if __getattribute__ is not None:
             setattr(type_, "__getattribute__", __getattribute__)
@@ -107,93 +126,103 @@ class AttrCleaner:
         return False
 
 
-class _Repr:
-    """
-    Helper for `fixlongargs()`: Returns the given value in `__repr__()`.
-    """
-
-    def __init__(self, value):
-        self.value = value
-
-    def __repr__(self):
-        return self.value
-
-    __str__ = __repr__
-
-
-def parsekeywordpairs(signature):
-    tokens = Python3Lexer().get_tokens(signature)
+def parsekeywordpairs(signature: str) -> Dict[str, str]:
     preamble = True
     stack = []
-    substack = []
+    substack: List[str] = []
     parendepth = 0
-    for token, value in tokens:
+    annotation = False
+    for token, value in Python3Lexer().get_tokens(signature):
         if preamble:
             if token is Token.Punctuation and value == "(":
+                # First "(" starts the list of arguments
                 preamble = False
             continue
 
         if token is Token.Punctuation:
-            if value in ("(", "{", "["):
+            if value in "({[":
                 parendepth += 1
-            elif value in (")", "}", "]"):
+            elif value in ")}]":
                 parendepth -= 1
-            elif value == ":" and parendepth == -1:
-                # End of signature reached
-                break
-            if (value == "," and parendepth == 0) or (
-                value == ")" and parendepth == -1
-            ):
+            elif value == ":":
+                if parendepth == -1:
+                    # End of signature reached
+                    break
+                elif parendepth == 0:
+                    # Start of type annotation
+                    annotation = True
+
+            if (value, parendepth) in ((",", 0), (")", -1)):
+                # End of current argument
                 stack.append(substack)
                 substack = []
+                # If type annotation didn't end before, it does now.
+                annotation = False
                 continue
+        elif token is Token.Operator and value == "=" and parendepth == 0:
+            # End of type annotation
+            annotation = False
 
-        if value and (parendepth > 0 or value.strip()):
+        if value and not annotation and (parendepth > 0 or value.strip()):
             substack.append(value)
 
     return {item[0]: "".join(item[2:]) for item in stack if len(item) >= 3}
 
 
-def fixlongargs(f, argspec):
+def _fix_default_values(f: Callable, argspec: ArgSpec) -> ArgSpec:
     """Functions taking default arguments that are references to other objects
-    whose str() is too big will cause breakage, so we swap out the object
-    itself with the name it was referenced with in the source by parsing the
-    source itself !"""
-    if argspec[3] is None:
+    will cause breakage, so we swap out the object itself with the name it was
+    referenced with in the source by parsing the source itself!"""
+
+    if argspec.defaults is None and argspec.kwonly_defaults is None:
         # No keyword args, no need to do anything
-        return
-    values = list(argspec[3])
-    if not values:
-        return
-    keys = argspec[0][-len(values) :]
+        return argspec
+
     try:
-        src = inspect.getsourcelines(f)
+        src, _ = inspect.getsourcelines(f)
     except (OSError, IndexError):
         # IndexError is raised in inspect.findsource(), can happen in
         # some situations. See issue #94.
-        return
-    signature = "".join(src[0])
-    kwparsed = parsekeywordpairs(signature)
+        return argspec
+    except TypeError:
+        # No source code is available, so replace the default values with what we have.
+        if argspec.defaults is not None:
+            argspec.defaults = [_Repr(str(value)) for value in argspec.defaults]
+        if argspec.kwonly_defaults is not None:
+            argspec.kwonly_defaults = {
+                key: _Repr(str(value))
+                for key, value in argspec.kwonly_defaults.items()
+            }
+        return argspec
 
-    for i, (key, value) in enumerate(zip(keys, values)):
-        if len(repr(value)) != len(kwparsed[key]):
+    kwparsed = parsekeywordpairs("".join(src))
+
+    if argspec.defaults is not None:
+        values = list(argspec.defaults)
+        keys = argspec.args[-len(values) :]
+        for i, key in enumerate(keys):
             values[i] = _Repr(kwparsed[key])
 
-    argspec[3] = values
+        argspec.defaults = values
+    if argspec.kwonly_defaults is not None:
+        for key in argspec.kwonly_defaults.keys():
+            argspec.kwonly_defaults[key] = _Repr(kwparsed[key])
+
+    return argspec
 
 
-getpydocspec_re = LazyReCompile(
+_getpydocspec_re = LazyReCompile(
     r"([a-zA-Z_][a-zA-Z0-9_]*?)\((.*?)\)", re.DOTALL
 )
 
 
-def getpydocspec(f, func):
+def _getpydocspec(f: Callable) -> Optional[ArgSpec]:
     try:
         argspec = pydoc.getdoc(f)
     except NameError:
         return None
 
-    s = getpydocspec_re.search(argspec)
+    s = _getpydocspec_re.search(argspec)
     if s is None:
         return None
 
@@ -204,7 +233,7 @@ def getpydocspec(f, func):
     defaults = []
     varargs = varkwargs = None
     kwonly_args = []
-    kwonly_defaults = dict()
+    kwonly_defaults = {}
     for arg in s.group(2).split(","):
         arg = arg.strip()
         if arg.startswith("**"):
@@ -219,18 +248,18 @@ def getpydocspec(f, func):
             if varargs is not None:
                 kwonly_args.append(arg)
                 if default:
-                    kwonly_defaults[arg] = default
+                    kwonly_defaults[arg] = _Repr(default)
             else:
                 args.append(arg)
                 if default:
-                    defaults.append(default)
+                    defaults.append(_Repr(default))
 
     return ArgSpec(
         args, varargs, varkwargs, defaults, kwonly_args, kwonly_defaults, None
     )
 
 
-def getfuncprops(func, f):
+def getfuncprops(func: str, f: Callable) -> Optional[FuncProps]:
     # Check if it's a real bound method or if it's implicitly calling __init__
     # (i.e. FooClass(...) and not FooClass.__init__(...) -- the former would
     # not take 'self', the latter would:
@@ -251,30 +280,28 @@ def getfuncprops(func, f):
         # '__init__' throws xmlrpclib.Fault (see #202)
         return None
     try:
-        argspec = get_argspec_from_signature(f)
-        fixlongargs(f, argspec)
-        if len(argspec) == 4:
-            argspec = argspec + [list(), dict(), None]
-        argspec = ArgSpec(*argspec)
-        fprops = FuncProps(func, argspec, is_bound_method)
+        argspec = _get_argspec_from_signature(f)
+        fprops = FuncProps(
+            func, _fix_default_values(f, argspec), is_bound_method
+        )
     except (TypeError, KeyError, ValueError):
-        argspec = getpydocspec(f, func)
-        if argspec is None:
+        argspec_pydoc = _getpydocspec(f)
+        if argspec_pydoc is None:
             return None
         if inspect.ismethoddescriptor(f):
-            argspec.args.insert(0, "obj")
-        fprops = FuncProps(func, argspec, is_bound_method)
+            argspec_pydoc.args.insert(0, "obj")
+        fprops = FuncProps(func, argspec_pydoc, is_bound_method)
     return fprops
 
 
-def is_eval_safe_name(string):
+def is_eval_safe_name(string: str) -> bool:
     return all(
         part.isidentifier() and not keyword.iskeyword(part)
         for part in string.split(".")
     )
 
 
-def get_argspec_from_signature(f):
+def _get_argspec_from_signature(f: Callable) -> ArgSpec:
     """Get callable signature from inspect.signature in argspec format.
 
     inspect.signature is a Python 3 only function that returns the signature of
@@ -284,20 +311,23 @@ def get_argspec_from_signature(f):
 
     """
     args = []
-    varargs = varkwargs = None
+    varargs = None
+    varkwargs = None
     defaults = []
     kwonly = []
     kwonly_defaults = {}
     annotations = {}
 
+    # We use signature here instead of getfullargspec as the latter also returns
+    # self and cls (for class methods).
     signature = inspect.signature(f)
     for parameter in signature.parameters.values():
-        if parameter.annotation is not inspect._empty:
+        if parameter.annotation is not parameter.empty:
             annotations[parameter.name] = parameter.annotation
 
         if parameter.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
             args.append(parameter.name)
-            if parameter.default is not inspect._empty:
+            if parameter.default is not parameter.empty:
                 defaults.append(parameter.default)
         elif parameter.kind == inspect.Parameter.POSITIONAL_ONLY:
             args.append(parameter.name)
@@ -309,35 +339,24 @@ def get_argspec_from_signature(f):
         elif parameter.kind == inspect.Parameter.VAR_KEYWORD:
             varkwargs = parameter.name
 
-    # inspect.getfullargspec returns None for 'defaults', 'kwonly_defaults' and
-    # 'annotations' if there are no values for them.
-    if not defaults:
-        defaults = None
-
-    if not kwonly_defaults:
-        kwonly_defaults = None
-
-    if not annotations:
-        annotations = None
-
-    return [
+    return ArgSpec(
         args,
         varargs,
         varkwargs,
-        defaults,
+        defaults if defaults else None,
         kwonly,
-        kwonly_defaults,
-        annotations,
-    ]
+        kwonly_defaults if kwonly_defaults else None,
+        annotations if annotations else None,
+    )
 
 
-get_encoding_line_re = LazyReCompile(r"^.*coding[:=]\s*([-\w.]+).*$")
+_get_encoding_line_re = LazyReCompile(r"^.*coding[:=]\s*([-\w.]+).*$")
 
 
-def get_encoding(obj):
+def get_encoding(obj) -> str:
     """Try to obtain encoding information of the source of an object."""
     for line in inspect.findsource(obj)[0][:2]:
-        m = get_encoding_line_re.search(line)
+        m = _get_encoding_line_re.search(line)
         if m:
             return m.group(1)
     return "utf8"
@@ -346,20 +365,23 @@ def get_encoding(obj):
 def get_encoding_file(fname: str) -> str:
     """Try to obtain encoding information from a Python source file."""
     with open(fname, encoding="ascii", errors="ignore") as f:
-        for unused in range(2):
+        for _ in range(2):
             line = f.readline()
-            match = get_encoding_line_re.search(line)
+            match = _get_encoding_line_re.search(line)
             if match:
                 return match.group(1)
     return "utf8"
 
 
-def getattr_safe(obj: Any, name: str):
-    """side effect free getattr (calls getattr_static)."""
+def getattr_safe(obj: Any, name: str) -> Any:
+    """Side effect free getattr (calls getattr_static)."""
     result = inspect.getattr_static(obj, name)
     # Slots are a MemberDescriptorType
     if isinstance(result, MemberDescriptorType):
         result = getattr(obj, name)
+    # classmethods are safe to access (see #966)
+    if isinstance(result, (classmethod, staticmethod)):
+        result = result.__get__(obj, obj)
     return result
 
 
@@ -369,8 +391,3 @@ def hasattr_safe(obj: Any, name: str) -> bool:
         return True
     except AttributeError:
         return False
-
-
-def get_source_unicode(obj):
-    """Returns a decoded source of object"""
-    return inspect.getsource(obj)
