@@ -5,13 +5,15 @@ import string
 import sys
 
 from contextlib import contextmanager
+from typing import cast
 from curtsies.formatstringarray import (
     fsarray,
     assertFSArraysEqual,
     assertFSArraysEqualIgnoringFormatting,
 )
 from curtsies.fmtfuncs import cyan, bold, green, yellow, on_magenta, red
-from unittest import mock
+from curtsies.window import CursorAwareWindow
+from unittest import mock, skipIf
 
 from bpython.curtsiesfrontend.events import RefreshRequestEvent
 from bpython import config, inspection
@@ -56,7 +58,7 @@ class CurtsiesPaintingTest(ClearEnviron):
             def _request_refresh(inner_self):
                 pass
 
-        self.repl = TestRepl(config=setup_config())
+        self.repl = TestRepl(setup_config(), cast(CursorAwareWindow, None))
         self.repl.height, self.repl.width = (5, 10)
 
     @property
@@ -284,7 +286,9 @@ class HigherLevelCurtsiesPaintingTest(CurtsiesPaintingTest):
             def _request_refresh(inner_self):
                 self.refresh()
 
-        self.repl = TestRepl(banner="", config=setup_config())
+        self.repl = TestRepl(
+            setup_config(), cast(CursorAwareWindow, None), banner=""
+        )
         self.repl.height, self.repl.width = (5, 32)
 
     def send_key(self, key):
@@ -307,6 +311,10 @@ class TestWidthAwareness(HigherLevelCurtsiesPaintingTest):
         cursor_pos = self.repl.paint()[1]
         self.assertEqual(cursor_pos, (1, 4))
 
+    @skipIf(
+        sys.version_info[:2] >= (3, 11) and sys.version_info[:3] < (3, 11, 1),
+        "https://github.com/python/cpython/issues/98744",
+    )
     def test_display_of_padding_chars(self):
         self.repl.width = 11
         [self.repl.add_normal_character(c) for c in "ｗｉｄｔｈ"]
