@@ -1,6 +1,9 @@
+import inspect
 import os
 import sys
 import unittest
+from collections.abc import Sequence
+from typing import List
 
 from bpython import inspection
 from bpython.test.fodder import encoding_ascii
@@ -8,6 +11,7 @@ from bpython.test.fodder import encoding_latin1
 from bpython.test.fodder import encoding_utf8
 
 pypy = "PyPy" in sys.version
+_is_py311 = sys.version_info[:2] >= (3, 11)
 
 try:
     import numpy
@@ -50,23 +54,17 @@ class TestInspection(unittest.TestCase):
         def fails(spam=["-a", "-b"]):
             pass
 
-        default_arg_repr = "['-a', '-b']"
-        self.assertEqual(
-            str(["-a", "-b"]),
-            default_arg_repr,
-            "This test is broken (repr does not match), fix me.",
-        )
-
         argspec = inspection.getfuncprops("fails", fails)
+        self.assertIsNotNone(argspec)
         defaults = argspec.argspec.defaults
-        self.assertEqual(str(defaults[0]), default_arg_repr)
+        self.assertEqual(str(defaults[0]), '["-a", "-b"]')
 
     def test_pasekeywordpairs_string(self):
         def spam(eggs="foo, bar"):
             pass
 
         defaults = inspection.getfuncprops("spam", spam).argspec.defaults
-        self.assertEqual(repr(defaults[0]), "'foo, bar'")
+        self.assertEqual(repr(defaults[0]), '"foo, bar"')
 
     def test_parsekeywordpairs_multiple_keywords(self):
         def spam(eggs=23, foobar="yay"):
@@ -74,7 +72,14 @@ class TestInspection(unittest.TestCase):
 
         defaults = inspection.getfuncprops("spam", spam).argspec.defaults
         self.assertEqual(repr(defaults[0]), "23")
-        self.assertEqual(repr(defaults[1]), "'yay'")
+        self.assertEqual(repr(defaults[1]), '"yay"')
+
+    def test_pasekeywordpairs_annotation(self):
+        def spam(eggs: str = "foo, bar"):
+            pass
+
+        defaults = inspection.getfuncprops("spam", spam).argspec.defaults
+        self.assertEqual(repr(defaults[0]), '"foo, bar"')
 
     def test_get_encoding_ascii(self):
         self.assertEqual(inspection.get_encoding(encoding_ascii), "ascii")
@@ -89,19 +94,13 @@ class TestInspection(unittest.TestCase):
         self.assertEqual(inspection.get_encoding(encoding_utf8.foo), "utf-8")
 
     def test_get_source_ascii(self):
-        self.assertEqual(
-            inspection.get_source_unicode(encoding_ascii.foo), foo_ascii_only
-        )
+        self.assertEqual(inspect.getsource(encoding_ascii.foo), foo_ascii_only)
 
     def test_get_source_utf8(self):
-        self.assertEqual(
-            inspection.get_source_unicode(encoding_utf8.foo), foo_non_ascii
-        )
+        self.assertEqual(inspect.getsource(encoding_utf8.foo), foo_non_ascii)
 
     def test_get_source_latin1(self):
-        self.assertEqual(
-            inspection.get_source_unicode(encoding_latin1.foo), foo_non_ascii
-        )
+        self.assertEqual(inspect.getsource(encoding_latin1.foo), foo_non_ascii)
 
     def test_get_source_file(self):
         path = os.path.join(
@@ -130,8 +129,15 @@ class TestInspection(unittest.TestCase):
         self.assertIn("file", props.argspec.kwonly)
         self.assertIn("flush", props.argspec.kwonly)
         self.assertIn("sep", props.argspec.kwonly)
-        self.assertEqual(props.argspec.kwonly_defaults["file"], "sys.stdout")
-        self.assertEqual(props.argspec.kwonly_defaults["flush"], "False")
+        if _is_py311:
+            self.assertEqual(
+                repr(props.argspec.kwonly_defaults["file"]), "None"
+            )
+        else:
+            self.assertEqual(
+                repr(props.argspec.kwonly_defaults["file"]), "sys.stdout"
+            )
+        self.assertEqual(repr(props.argspec.kwonly_defaults["flush"]), "False")
 
     @unittest.skipUnless(
         numpy is not None and numpy.__version__ >= "1.18",
@@ -145,6 +151,111 @@ class TestInspection(unittest.TestCase):
         # np.array(object, dtype=None, *, ...).
         self.assertEqual(props.argspec.args, ["object", "dtype"])
 
+    def test_issue_966_freestanding(self):
+        def fun(number, lst=[]):
+            """
+            Return a list of numbers
+
+            Example:
+            ========
+            C.cmethod(1337, [1, 2]) # => [1, 2, 1337]
+            """
+            return lst + [number]
+
+        def fun_annotations(number: int, lst: List[int] = []) -> List[int]:
+            """
+            Return a list of numbers
+
+            Example:
+            ========
+            C.cmethod(1337, [1, 2]) # => [1, 2, 1337]
+            """
+            return lst + [number]
+
+        props = inspection.getfuncprops("fun", fun)
+        self.assertEqual(props.func, "fun")
+        self.assertEqual(props.argspec.args, ["number", "lst"])
+        self.assertEqual(repr(props.argspec.defaults[0]), "[]")
+
+        props = inspection.getfuncprops("fun_annotations", fun_annotations)
+        self.assertEqual(props.func, "fun_annotations")
+        self.assertEqual(props.argspec.args, ["number", "lst"])
+        self.assertEqual(repr(props.argspec.defaults[0]), "[]")
+
+    def test_issue_966_class_method(self):
+        class Issue966(Sequence):
+            @classmethod
+            def cmethod(cls, number: int, lst: List[int] = []):
+                """
+                Return a list of numbers
+
+                Example:
+                ========
+                C.cmethod(1337, [1, 2]) # => [1, 2, 1337]
+                """
+                return lst + [number]
+
+            @classmethod
+            def bmethod(cls, number, lst):
+                """
+                Return a list of numbers
+
+                Example:
+                ========
+                C.cmethod(1337, [1, 2]) # => [1, 2, 1337]
+                """
+                return lst + [number]
+
+        props = inspection.getfuncprops(
+            "bmethod", inspection.getattr_safe(Issue966, "bmethod")
+        )
+        self.assertEqual(props.func, "bmethod")
+        self.assertEqual(props.argspec.args, ["number", "lst"])
+
+        props = inspection.getfuncprops(
+            "cmethod", inspection.getattr_safe(Issue966, "cmethod")
+        )
+        self.assertEqual(props.func, "cmethod")
+        self.assertEqual(props.argspec.args, ["number", "lst"])
+        self.assertEqual(repr(props.argspec.defaults[0]), "[]")
+
+    def test_issue_966_static_method(self):
+        class Issue966(Sequence):
+            @staticmethod
+            def cmethod(number: int, lst: List[int] = []):
+                """
+                Return a list of numbers
+
+                Example:
+                ========
+                C.cmethod(1337, [1, 2]) # => [1, 2, 1337]
+                """
+                return lst + [number]
+
+            @staticmethod
+            def bmethod(number, lst):
+                """
+                Return a list of numbers
+
+                Example:
+                ========
+                C.cmethod(1337, [1, 2]) # => [1, 2, 1337]
+                """
+                return lst + [number]
+
+        props = inspection.getfuncprops(
+            "bmethod", inspection.getattr_safe(Issue966, "bmethod")
+        )
+        self.assertEqual(props.func, "bmethod")
+        self.assertEqual(props.argspec.args, ["number", "lst"])
+
+        props = inspection.getfuncprops(
+            "cmethod", inspection.getattr_safe(Issue966, "cmethod")
+        )
+        self.assertEqual(props.func, "cmethod")
+        self.assertEqual(props.argspec.args, ["number", "lst"])
+        self.assertEqual(repr(props.argspec.defaults[0]), "[]")
+
 
 class A:
     a = "a"
